@@ -13,6 +13,8 @@
  * 2000/12/02 - EvB - Created
  * 2001/02/06 - EvB - Made dbg_cvtstr NULL-proof
  * 2001/11/01 - EvB - Added standard logging functions
+ * 2002/02/06 - EvB - Added msg_line interface and made meta_printav
+ * 		      and hexdump use it instead of writing to stderr
  */
 
 char debug_id[] = "DEBUG - Copyright (C) 2000 Emile van Bergen.";
@@ -94,6 +96,10 @@ struct {
 	{ 0,		-1 },		/* not found */
 };
 
+#ifndef LOG_MAKEPRI
+#define LOG_MAKEPRI(f, p)	(((f) << 3) | (p))
+#endif
+
 #endif
 
 
@@ -178,16 +184,22 @@ void msg(int ourfac, int level, char *fmt, ...)
 #else
 	vsprintf(logline + n, fmt, ap);		/* Cross your fingers */
 #endif
+	msg_line(level, logline);
+	va_end(ap);
+}
 
+
+void msg_line(int level, char *logline)
+{
 	/* Write line to stderr if we haven't been told to be quiet */
-	if (!msg_nostderr) write(2, logline, strlen(logline));
+	if (!msg_nostderr) 
+		write(2, logline, strlen(logline));
 
 #if HAVE_SYSLOG
 	/* Write line to syslog if facility is not none */
-	if (msg_syslogfac != -1) syslog(level, "%s", logline);
+	if (msg_syslogfac != -1) 
+		syslog(LOG_MAKEPRI(msg_syslogfac, level), "%s", logline);
 #endif
-
-	va_end(ap);
 }
 
 
@@ -203,7 +215,7 @@ char *dbg_cvtstr(char *s, int slen)
 }
 
 
-#define DBG_HEXBUFLEN	(4 + 2 + 16 * 3 + 1 + 16 + 1)
+#define DBG_HEXBUFLEN	(4 + 2 + 16 * 3 + 1 + 16 + 1 + 1)
 
 void hexdump(char *s, int slen)
 {
@@ -213,7 +225,7 @@ void hexdump(char *s, int slen)
 
 	for(ofs = 0; ofs < slen; ofs += 16, s += 16) {
 		o = hexbuf;
-		memset(o, ' ', DBG_HEXBUFLEN);
+		memset(o, ' ', DBG_HEXBUFLEN - 1);
 		o += meta_ordtoa(o, 4, 4, 16, ofs) + 2;
 		for(i = 0; i < MIN(slen - ofs, 16); i++) 
 			o += meta_ordtoa(o, 2, 2, 16, s[i] & 0xff) + 1;
@@ -221,7 +233,8 @@ void hexdump(char *s, int slen)
 		for(i = 0; i < MIN(slen - ofs, 16); i++) 
 			*o++ = s[i] > 31 && s[i] < 127 ? s[i] : '.';
 		hexbuf[4 + 2 + 16 * 3 + 1 + 16] = '\n';
-		write(2, hexbuf, DBG_HEXBUFLEN);
+		hexbuf[DBG_HEXBUFLEN - 1] = 0;
+		msg_line(L_DEBUG, hexbuf);
 	}
 }
 
