@@ -23,6 +23,23 @@
  * 2001/09/11 - EvB - Added meta_addav here; used by job_new, chan_handle_read
  *		      and vm_run (we should, at least).
  * 2001/09/14 - EvB - Added ACL support to meta_avtomsg here after some doubt.
+ * 2001/12/27 - EvB - Rewrote chan_*msgtoav into meta_*msgtoav here.
+ * 2002/02/04 - EvB - Fixed meta_decode to allow zero-sized encapsulating 
+ * 		      attributes
+ * 		    - Added more diagnostics output to meta_decode
+ * 		    - Made negative len_adj values for zero-sized length fields
+ * 		      to signify a (skip) length relative to the enclosing
+ * 		      block. Useful for decoding relative-sized fixed fields 
+ * 		      without accompanying length fields.
+ *		    - Meta_decode searches the dictionary 1. for an attribute 
+ *		      that has the right ID and vendor PEC, 2. for an attribute 
+ *		      that has the right ID and vendor=any, 3. for an attribute
+ *		      that has the right vendor and ID=any, 4. for an attribute
+ *		      that has vendor=any and ID=any. Step 3 used to be left
+ *		      out.
+ * 2002/03/01 - EvB - Made buildtree even more gross by having it delete
+ * 		      items with noenc=1 from its source list. Removes memory
+ * 		      leak though. Cleanup still needed.
  */
 
 char metaops_id[] = "METAOPS - Copyright (C) 2001 Emile van Bergen.";
@@ -34,7 +51,7 @@ char metaops_id[] = "METAOPS - Copyright (C) 2001 Emile van Bergen.";
 
 
 #include <unistd.h>	/* For the write() in printav */
-#include <malloc.h>
+#include <stdlib.h>
 #include <string.h>
 
 #include <misc.h>	/* For hex() */
@@ -52,24 +69,115 @@ char metaops_id[] = "METAOPS - Copyright (C) 2001 Emile van Bergen.";
  */
 
 
+
+/*
+ * Helper macros and functions for meta_decode
+ */
+
+
 /* Return the attribute or vendor field */
 
 #define getatrnr(s, d)	(getord((char *)(d) + (s)->atr_ofs, (s)->atr_size))
 #define getvndnr(s, d)	(getord((char *)(d) + (s)->vnd_ofs, (s)->vnd_size))
 
-/* Return the length field + the adjustment if the size of the length field 
-   is non-zero, otherwise return the adjustment itself, as an absolute skip 
-   length */
+/* The so-called skip length is the length needed to get from offset 0 in this
+   attribute to the next attribute. The value length is the length of the 
+   value itself. 
 
-#define getskiplen(i, d)	\
-	((i)->len_adj + getord((char *)(d) + (i)->len_ofs, (i)->len_size))
+   The skip length can be obtained as follows: if the item has a length field,
+   (detected by len_size being non-zero), take its value and add the adjustment
+   given by len_adj, which can be positive, negative or zero.
 
-/* Return the given skiplength + the item's value size adjustment if that's
-   zero or negative, otherwise return the item's value size itself as the 
-   absolute size. */
+   If there is no length field, and len_adj is zero or less, take the size
+   of the enclosing block and add len_adj. If there is no length field but
+   len_adj is positive, then skiplen is specified by the absolute value of
+   len_adj.
 
-#define getvallen(i, sl)	\
-	((i)->val_size + ((i)->val_size <= 0 ? (sl) : (META_ORD)0))
+   The value length is calculated like this: if val_size is zero or less, take 
+   skiplen and add val_size. If val_size is positive, then the value length
+   is defined by the absolute value of val_size.
+   
+   This scheme allows for any combination of absolute or relative
+   specifications for the offset of the next attribute and for the size of the
+   value itself. */
+
+#define getskiplen(i, d, bl)						      \
+	((i)->len_adj + ( (i)->len_size ?				      \
+			  getord((char *)(d) + (i)->len_ofs, (i)->len_size) : \
+			  ( (i)->len_adj <= 0 ? 			      \
+			    (bl) : 0)))
+
+#define getvallen(i, sl)						      \
+	((i)->val_size + ( (i)->val_size <= 0 ? 			      \
+			   (sl) : 0))
+
+
+/* Post-mortem decoding diagnostics */
+
+static void debug_fixed(META_ITEM *i, char *blk, META_ORD blksize)
+{
+	if (i->len_ofs + i->len_size > blksize) {
+		msg(F_RECV, L_DEBUG, "meta_decode: ERROR: Length field for "
+				     "%s/%ld/%s at 0x%04lx, length 0x%lx not "
+				     "within area, length 0x%04lx!\n", 
+		    i->spc->name, i->vnd, i->name,
+		    i->len_ofs, i->len_size, blksize);
+	} 
+	else {
+		msg(F_RECV, L_DEBUG, "meta_decode: ERROR: Field %s/%ld/%s at "
+				     "0x%04lx, length 0x%lx not within area, "
+				     "length 0x%04lx!\n",
+		    i->spc->name, i->vnd, i->name,
+		    i->val_ofs, getvallen(i, getskiplen(i, blk, blksize)), 
+		    blksize);
+	}
+}
+
+
+static void debug_attr(META_SPC *s, META_ITEM *i, char *blk, char *p, 
+		       META_ORD blksize, META_ORD defvnd)
+{
+	META_ORD atr, vnd, skiplen, vallen;
+
+	if (p > blk + blksize) {
+		msg(F_RECV, L_DEBUG, "meta_decode: ERROR: Arrived at 0x%04lx "
+				     "in space %s which is beyond 0x%04lx!\n",
+		    p - blk, s->name, blksize);
+		return;
+	}
+
+	if (p + s->atr_ofs + s->atr_size > blk + blksize) {
+		msg(F_RECV, L_DEBUG, "meta_decode: ERROR: Attribute field in "
+				     "space %s at 0x%04lx, length 0x%lx not "
+				     "within area, length 0x%04lx!\n",
+		    s->name, p - blk + s->atr_ofs, s->atr_size, blksize);
+		return;
+	}
+
+	atr = getatrnr(s, p);
+	if (p + s->vnd_ofs + s->vnd_size > blk + blksize) {
+		msg(F_RECV, L_DEBUG, "meta_decode: ERROR: Vendor field in "
+				     "space %s at 0x%04lx, length 0x%lx not "
+				     "within area, length 0x%04lx!\n",
+		    s->name, p - blk + s->vnd_ofs, s->vnd_size, blksize);
+		return;
+	}
+
+	vnd = s->vnd_size ? getvndnr(s, p) : defvnd;
+	if (!i) {
+		msg(F_RECV, L_DEBUG, "meta_decode: ERROR: Space %s has no attr"
+				     "ibute that matches id %ld, vendor %ld!\n",
+		    s->name, atr, vnd);
+		return;
+	}
+
+	skiplen = getskiplen(i, p, blksize);
+	vallen = getvallen(i, skiplen); 
+	msg(F_RECV, L_DEBUG, "meta_decode: ERROR: Value for attribute %s/%ld/%s"
+			     " at 0x%04lx, length 0x%lx (of 0x%lx) not within "
+			     "area, length 0x%04lx!\n",
+	    s->name, vnd, atr, p - blk + i->val_ofs, vallen, skiplen, blksize);
+}
 
 
 /* Decodes a data block, forming a flat list of AV items. */
@@ -94,10 +202,19 @@ META_AV *meta_decode(META *m, META_SPC *s, META_ORD defvnd,
  	/* Check if this is an A/V or fixed field space */	
 	if (s->atr_size == 0) {
 
+		/* Test here if the space doesn't have any items and tell about
+		   it. Otherwise we'd also return 0 but without telling why. */
+		if (!s->items) {
+			msg(F_RECV, L_DEBUG, "meta_decode: ERROR: Space %s "
+					     "does not contain any fields!\n",
+			    s->name);
+			return 0;
+		}
+
 		/* Fixed fields - loop through the items. */
 		for(i = s->items; 
 		    i && i->len_ofs + i->len_size <= blksize && 
-		    	 i->val_ofs + (vallen = getvallen(i, getskiplen(i, blk))) <= blksize; 
+		    	 i->val_ofs + (vallen = getvallen(i, getskiplen(i, blk, blksize))) <= blksize; 
 		    i = i->next) {
 
 			/* Add the data, unless we shouldn't decode it */
@@ -118,8 +235,8 @@ META_AV *meta_decode(META *m, META_SPC *s, META_ORD defvnd,
 				tail = av;
 			}
 
-			/* Always recurse into a subspace, if any */
-			if (i->subspace) {
+			/* Recurse into a subspace, if any */
+			if (i->subspace && vallen) {
 
 				av = meta_decode(m, i->subspace, defvnd,
 						 blk + i->val_ofs, vallen,
@@ -133,8 +250,9 @@ META_AV *meta_decode(META *m, META_SPC *s, META_ORD defvnd,
 			}
 		}
 
-		/* If we terminated prematurely, return -1 to signal an error */
-		if (i) { meta_freeavlist(avlist); return 0; }
+		/* If we terminated prematurely, show error and return */
+		if (i) { debug_fixed(i, blk, blksize); meta_freeavlist(avlist); 
+			 return 0; }
 
 		/* Return the head of the created list */
 		if (rettail) *rettail = tail;
@@ -147,16 +265,17 @@ META_AV *meta_decode(META *m, META_SPC *s, META_ORD defvnd,
 	skiplen = 0;
 
 	/* Attribute space. You've just got to love C's logical operators... */
-	for(p = blk;
+	for(p = blk, i = 0;
 	    p < blk + blksize &&
 	    	 p + s->atr_ofs + s->atr_size <= blk + blksize &&
 	    	 p + s->vnd_ofs + s->vnd_size <= blk + blksize &&
 		 ((i = meta_getitembynr(m, s, (atrnr = getatrnr(s, p)),
 		 	(vndnr = (s->vnd_size ? getvndnr(s, p) : defvnd)))) ||
 		  (i = meta_getitembynr(m, s, atrnr, META_ORD_ERR)) ||
+		  (i = meta_getitembynr(m, s, META_ORD_ERR, vndnr)) ||
 		  (i = meta_getitembynr(m, s, META_ORD_ERR, META_ORD_ERR))) &&
-	    	 p + i->val_ofs + ((skiplen = getskiplen(i, p)), (vallen = getvallen(i, skiplen))) <= blk + blksize;
-	    p += skiplen) {
+	    	 p + i->val_ofs + ((skiplen = getskiplen(i, p, blksize)), (vallen = getvallen(i, skiplen))) <= blk + blksize;
+	    p += skiplen, i = 0) {
 
 		/* Add the data, unless we shouldn't decode it */
 		if (!i->nodec) {
@@ -177,7 +296,7 @@ META_AV *meta_decode(META *m, META_SPC *s, META_ORD defvnd,
 		}
 
 		/* Recurse into a subspace, if any */
-		if (i->subspace) {
+		if (i->subspace && vallen) {
 			av = meta_decode(m, i->subspace, vndnr,
 					 p + i->val_ofs, vallen, &newtail);
 			if (!av) { meta_freeavlist(avlist); return 0; }
@@ -189,8 +308,12 @@ META_AV *meta_decode(META *m, META_SPC *s, META_ORD defvnd,
 		}
 	}
 
-	/* If the loop terminated prematurely, return 0 to signal the error */
-	if (p != blk + blksize) { meta_freeavlist(avlist); return 0; }
+	/* If the loop terminated prematurely, show and return 0 to signal it */
+	if (p != blk + blksize) { 
+		debug_attr(s, i, blk, p, blksize, defvnd); 
+		meta_freeavlist(avlist); 
+		return 0; 
+	}
 
 	/* Return the (tail and) head of the created list */
 	if (rettail) *rettail = tail;
@@ -223,7 +346,7 @@ META_ORD meta_encode(META_SPC *s, char *blk, META_ORD blksize, META_AV *avlist,
 
 	    	/* Skip deleted items. Don't skip items with 'noenc' specified,
 		   they're only skipped in buildtree so they can still be used
-		   as encapsulating items. */
+		   as encapsulation items. */
 		if (av->flags & AV_DEL) continue;
 
 		/* Check if the vendor and attribute fields fit */
@@ -318,7 +441,7 @@ META_ORD meta_encode(META_SPC *s, char *blk, META_ORD blksize, META_AV *avlist,
 
 /* Build a tree from a list, inserting encapsulating attributes where needed.
    It's become a bit of an ugly routine, after adding the double-linked AV
-   list and the AV_DEL flag & noenc handling. */
+   list and the AV_DEL flag & noenc handling. Should be cleaned up - I know. */
 
 void meta_buildtree(META_AV **srclst, META_AV **dstlst, META_SPC *groundspc)
 {
@@ -335,15 +458,14 @@ void meta_buildtree(META_AV **srclst, META_AV **dstlst, META_SPC *groundspc)
 	itmsp = dstsp = 0; 
 	curspc = groundspc;
 	
-skipdel_0:
-	/* Take the first AV from the source list. We _move_ items here. */
-	curav = *srclst;
+	/* Find first non-deleted item */
+	for(curav = *srclst; 
+	    curav && ((curav->flags & AV_DEL) || (curav->i && curav->i->noenc));
+	    curav = curav->next);
 	if (!curav) { *dstlst = 0; return; }
-	*srclst = curav->next;
-	curav->next = 0;
-	curav->prev = 0;
-	if (curav->flags & AV_DEL || (curav->i && curav->i->noenc))
-		goto skipdel_0;
+
+	/* Remove it from the list */
+	meta_remav(srclst, 0, curav);
 
 	for(;;) {
 
@@ -351,32 +473,20 @@ skipdel_0:
 		if (curav->i->spc == curspc) {
 
 			/* Yes, so add the item to the destination list */
-			*dstlst = curav;
-			dstlst = &(curav->next);
+			*dstlst = curav; 
+			dstlst = &(curav->next); 
 
-skipdel_1:		/* Take the next item from the source list and cont. */
-			curav = *srclst;
-			if (curav) {
+			/* Skip deleted items and take next one from the list */
+			do { curav = curav->next; } while(curav && ((curav->flags & AV_DEL) || (curav->i && curav->i->noenc)));
+			*dstlst = 0; if (!curav) return;
+			meta_remav(srclst, 0, curav);
 
-				/* There actually is a next item */
-				*srclst = curav->next;
-				curav->next = 0;
-				curav->prev = 0;
-				if (curav->flags & AV_DEL || 
-				    (curav->i && curav->i->noenc))
-					goto skipdel_1;
-
-				/* If this is a single attr. space, leave
-				   the space now. */
-				while(curspc->single && dstsp) {
-					curspc = spcstk[--dstsp];
-					dstlst = dststk[dstsp];
-				}
-				continue;
+			/* If this is a single attr. space, leave it now. */
+			while(curspc->single && dstsp) {
+				curspc = spcstk[--dstsp];
+				dstlst = dststk[dstsp];
 			}
-			
-			/* Nothing left on the source list, so we're done. */
-			return;
+			continue;
 		}
 
 		/* No, so try to reach the current space from this item's space
@@ -478,7 +588,256 @@ skipdel_1:		/* Take the next item from the source list and cont. */
 }
 
 
-/* Build a message holding an AV list data in binary or ASCII form */
+/* Here's the deal. The *msgtoav() functions get a buffer, not a ring,
+   so that the radius client module can read() into an aligned buffer
+   instead of into a ring which is then emptied to the aligned buffer.
+   And, we have to copy before doing getitembyspec() etc. anyway.
+
+   The ASCII and binary functions stay separate. There's no need to
+   virtualise them here as there would still be a difference in
+   semantics (call per message part or per full message) of which the
+   upper layer must be aware.
+
+   I'm not touching avtomsg* right now. Agreed, the binary version could be
+   faster, especially considering the fact that we WILL NOT make these 
+   functions output straight to a ring, also because both printav() AND the 
+   radius client only require a buffer, not a ring.
+
+   Instead, a case to split avtomsg into avtobinmsg and avtoascmsg could
+   be made: the only place where the choice must then be made dependent on
+   the interface's flags is in job_tochan(). But the case for separation
+   is not as strong as for *msgtoav(), because this always converts a full
+   list to a full message, regardless of the interface type. */
+
+
+/* Add one AV pair from an ASCII message to an AV list. Returns -1 for
+   all errors. WARNING: the parameter 'buf' is modified. */
+
+int meta_ascmsgtoav(META *m, char *buf, ssize_t len,
+		    META_AV **head, META_AV **tail,
+		    int flags, META_AV *acl)
+{
+	char *end, *atr, *c, *o;
+	META_AV *av, *aclav;
+	META_VAL *v;
+	ssize_t rl;
+
+	/* Init av for easy cleanup and set end past end of partial message */
+	av = 0;
+	end = buf + len;
+
+	/* Skip initial spaces and tabs */
+	for(atr = buf; atr < end && (*atr == ' ' || *atr == '\t'); atr++);
+
+	/* End the attr at the first space, tab or equals sign */
+	for(c = atr; c < end && *c != ' ' && *c != '\t' && *c != '='; c++);
+	if (c >= end) goto meta_ascmsgtoav_invpair;
+	*c = 0;
+
+	/* Skip all following spaces, tabs and equals signs (Jon P. says: 
+	   be liberal in what you accept). After this, atr is the zero 
+	   terminated attribute spec and c..end is the raw value. */
+	for(c++; c < end && (*c == ' ' || *c == '\t' || *c == '='); c++);
+
+	/* Allocate new A/V item at this point */
+	av = (META_AV *)malloc(sizeof(META_AV));
+	if (!av) { msg(F_MISC, L_ERR, "meta_ascmsgtoav: ERROR: No memory!\n"); 
+		   return -1; }
+	memset(av, 0, sizeof(META_AV));
+
+	/* Find the specified item in the dictionary */
+	msg(F_PROC, L_DEBUG, "- got attribute '%s'\n", atr);
+	av->i = meta_getitembyspec(m, atr);
+	if (!av->i) { msg(F_PROC, L_ERR, "meta_ascmsgtoav: ERROR: Unknown "
+					 "attribute '%s'!\n", atr);
+		      free(av); return -1; }
+
+	/* Check if the item is allowed if we have an ACL */
+	if (acl) {
+		for(aclav = acl; 
+		    aclav && aclav->i != av->i; 
+		    aclav = aclav->next);
+
+		/* Return with no error if not found in ACL */
+		if (!aclav) { 
+			msg(F_PROC, L_DEBUG, "  not in ACL; ignored\n");
+			free(av);
+			return 0;
+		}
+	}
+
+	/* Now parse the value; first check its style */
+	if (flags & AVMSG_HEXVALUE) {
+
+		/* Hex values for all data types */
+		if (MT_ISORD(av->i->val_type)) {
+
+			/* Ordinal type; do not allow empty values */
+			if (c >= end) goto meta_ascmsgtoav_invpair;
+			av->l = meta_atoord(c, end - c, 0, 0, &rl, 16);
+			if (!rl) goto meta_ascmsgtoav_invpair;
+		}
+		else if (c < end) {
+
+			/* Non-empty string type. Allocate half the number of 
+			   input bytes, rounding up. Is always enough. */
+			av->p = (char *)malloc((end - c + 1) >> 1);
+			if (!av->p) { msg(F_MISC, L_ERR, "meta_ascmsgtoav: "
+							 "ERROR: No memory!\n");
+				      free(av); return -1; }
+
+			/* Fill the string, one byte out per two bytes in */
+			for(o = av->p; end - c >= 2; c++, c++, o++) {
+				*o = meta_atoord(c, 2, 0, 0, &rl, 16);
+				if (rl != 2) break;
+			}
+			
+			/* Set the length and flags */
+			av->l = o - av->p;
+			av->flags |= AV_FREE_P;
+		}
+
+		/* Add pair to list and return. */
+		meta_addav(head, tail, 0, 0, av);
+		return 0;
+	}
+
+	/* No hex values; do not allow empty values */
+	if (c >= end) goto meta_ascmsgtoav_invpair;
+
+	/* Parse value according to item's data type */
+	switch(av->i->val_type) {
+	  case MT_INTEGER:
+	  case MT_DATE:
+		av->l = meta_atoord(c, end - c, 0, 0, &rl, 10);
+		if (rl) break;
+
+		/* Try to find named constant */
+		*end = 0;			/* allowed, see top */
+		v = meta_getvalbyname(m, av->i, c);
+		if (!v) goto meta_ascmsgtoav_invpair;
+		av->l = v->nr;
+		break;
+
+	  case MT_IPADDR:
+		av->l = meta_atoip(c, end - c, 0, 0, &rl);
+		if (!rl) goto meta_ascmsgtoav_invpair;
+		break;
+
+	  default:
+	  	/* Note: we allocate as many bytes as there are left in the
+		   input message. This is safe, because prttoa can never 
+		   output more data than we feed it, only less. Guaranteed. */
+		av->p = (char *)malloc(end - c);
+		if (!av->p) { msg(F_MISC, L_ERR, "neta_ascmsgtoav: ERROR: No "
+						 "memory!\n"); 
+			      free(av); return -1; }
+		av->l = meta_prttoa(c, end - c, 0, 0, &rl, av->p, end - c);
+		av->flags |= AV_FREE_P;
+	}
+
+	/* Add pair to list and return. */
+	meta_addav(head, tail, 0, 0, av);
+	return 0;
+
+meta_ascmsgtoav_invpair:
+	msg(F_PROC, L_NOTICE, "meta_ascmsgtoav: ERROR: Invalid AV pair '%s'!\n",
+	    dbg_cvtstr(buf, len));
+	if (av) free(av);
+	return -1;
+}
+
+
+/* Add one or more AV pairs from a binary message to an AV list */
+
+int meta_binmsgtoav(META *m, U_INT32_T *buf, ssize_t len,
+		    META_AV **head, META_AV **tail,
+		    int flags, META_AV *acl)
+{
+	U_INT32_T *i, *e;
+	META_ORD spcnr, vndnr, atrnr;
+	ssize_t atrlen;
+	META_SPC *spc;
+	META_AV *av, *aclav;
+
+	/* Loop through attributes, starting at (char *)msg + 8 */
+	e = buf + ((len + 3) >> 2);
+	for(i = buf + 2; i < e; i += (atrlen + 3) >> 2) {
+
+		/* Get values from attribute header. Warning: netint32 will
+		   evaluate its argument 4 times if little-endian ;o) */
+		spcnr = netint32(*i); i++;
+		vndnr = netint32(*i); i++;
+		atrnr = netint32(*i); i++;
+		atrlen = netint32(*i); i++;
+		msg(F_PROC, L_DEBUG, "- got spcnr %ld, vndnr %ld, atrnr %ld, "
+				     "atrlen %ld\n",
+		    (long)spcnr, (long)vndnr, (long)atrnr, (long)atrlen);
+
+		spc = meta_getspcbynr(m, spcnr);
+		if (!spc) { msg(F_PROC, L_NOTICE, "meta_binmsgtoav: WARNING: "
+						  "Unknown space %d!\n", spcnr);
+			    continue; }
+
+		/* Allocate new A/V item at this point */
+		av = (META_AV *)malloc(sizeof(META_AV));
+		if (!av) { msg(F_MISC, L_ERR, "meta_binmsgtoav: ERROR: No "
+					      "memory!\n"); return -1; }
+		memset(av, 0, sizeof(META_AV));
+
+		/* Find the item in the dictionary */
+		av->i = meta_getitembynr(m, spc, atrnr, vndnr);
+		if (!av->i) { msg(F_PROC, L_NOTICE, "meta_binmsgtoav: WARNING: "
+						    "Unknown attribute %d, "
+						    "vendor %d, space %d!\n",
+				  atrnr, vndnr, spcnr);
+			      free(av); continue; }
+
+		/* See if we have a non-empty receive ACL */
+		if (acl) {
+
+			/* We do - find attribute in ACL */
+			for(aclav = acl;
+			    aclav && aclav->i != av->i;
+			    aclav = aclav->next);
+		    
+			/* Skip attribute if not found */
+			if (!aclav) {
+				msg(F_PROC, L_DEBUG, "  not in ACL; ignored\n");
+				free(av);
+				continue;
+			}
+		}
+
+		/* Get data according to type */
+		if (MT_ISORD(av->i->val_type)) {
+
+			/* Ordinal - get value from even-multiple-of-4 sized
+			   field and store in av->l */
+			av->p = 0;
+			av->l = getord((char *)i, (atrlen + 3) & ~3);
+		}
+		else {
+			/* String - copy to new buf at av->p, length in av->l */
+			av->p = (char *)malloc(atrlen);
+			if (!av->p) { msg(F_MISC, L_ERR, "meta_binmsgtoav: "
+							 "ERROR: No memory!\n");
+				      free(av); return -1; }
+			memcpy(av->p, (char *)i, atrlen);
+			av->l = atrlen;
+			av->flags |= AV_FREE_P;
+		}
+
+		/* Add A/V item to list */
+		if (msg_thresh[F_PROC] >= L_DEBUG) meta_printav(m, av, 0);
+		meta_addav(head, tail, 0, 0, av);
+	}
+
+	return 0;
+}
+
+
+/* Build a message holding an AV list in binary or ASCII form */
 
 ssize_t meta_avtomsg(META *m, META_AV *avlst, 
 		     char *buf, ssize_t buflen, 
@@ -775,6 +1134,20 @@ void meta_addav(META_AV **head, META_AV **tail, META_AV *rel, int before,
 }
 
 
+/* Remove an AV item from a list, without freeing it */
+
+void meta_remav(META_AV **head, META_AV **tail, META_AV *av)
+{
+	/* Change next and previous item's linkage */
+	if (av->next) av->next->prev = av->prev;
+	if (av->prev) av->prev->next = av->next;
+
+	/* Adjust list's head or tail (optional) if necessary */
+	if (av == *head) *head = av->next;
+	if (tail && av == *tail) *tail = av->prev;
+}
+
+
 /* Free temporary data that may be held by an AV item */
 
 void meta_freeavdata(META_AV *av)
@@ -815,16 +1188,17 @@ void meta_printav(META *m, META_AV *av, int depth)
 	buflen = sizeof(buf);
 	ret = 0;
 	o = buf;
-	for(n = 0; n < depth && buflen > 0; n++) {
+	for(n = 0; n < depth && buflen > 1; n++) {
 		*o++ = '\t'; *o++ ='|';
 		buflen -= 2; ret += 2;
 	}
 
-	n = meta_avtomsg(m, av, o, buflen, AVMSG_ONESHOT + AVMSG_ASCII + 
+	n = meta_avtomsg(m, av, o, buflen - 1, AVMSG_ONESHOT + AVMSG_ASCII + 
 			 AVMSG_ADDTAB + AVMSG_ADDSPACES + AVMSG_NAMEDCONST, 0);
 	buflen -= n; ret += n;
 
-	write(2, buf, ret);
+	buf[ret] = 0; 
+	msg_line(L_DEBUG, buf);
 }
 
 
