@@ -26,10 +26,17 @@
 
 
 #include <sys/types.h>		/* for ssize_t */
+#include <sys/param.h>		/* for endianness macros */
+
+#include <metadata.h>		/* for META_ORD */
 
 
+#ifndef MIN
 #define MIN(a, b)	((a) < (b) ? (a) : (b))
+#endif
+#ifndef MAX
 #define MAX(a, b)	((a) > (b) ? (a) : (b))
+#endif
 
 
 /* Try to establish endianness - supports BSD, GNU and SysV style */
@@ -81,6 +88,8 @@
 char *memrchr(char *s, int c, ssize_t len);
 void hex(char *buf, const char *src, ssize_t len);
 
+void encrypt_attr_style_1(char *text, META_ORD *len,char *secret,char *reqauth);
+int decrypt_attr_style_1(char *text, META_ORD *len, char *secret,char *reqauth);
 
 #endif
 
