@@ -1,5 +1,5 @@
 
-#include <malloc.h>
+#include <stdlib.h>
 #include <unistd.h>
 
 #include <metadict.h>
