@@ -7,7 +7,7 @@
 #include <sys/time.h>
 #include <sys/wait.h>
 #include <time.h>
-#include <malloc.h>
+#include <stdlib.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <signal.h>
@@ -94,7 +94,7 @@ int main()
 
 	/* Create the unstarted subprocess and add to set */
 	p = proc_new("newproccld_asc arg1 arg2 arg3", 0, 
-		     C_DV_FLAGS_ASCII, 10, 0, 0);
+		     C_DV_FLAGS_ASCII, 10, 0, 0, 0);
 	if (!p) {
 		msg(F_MISC, L_ERR, "main: ERROR: Could not define subproc!\n");
 		free(ps); return 1;
