@@ -76,7 +76,7 @@ char radldap_id[] = "RADLDAP - Copyright (C) 2001 Emile van Bergen.";
 
 
 #ifndef RADLDAP_MAPFILE
-#define RADLDAP_MAPFILE	"radldap.attrmap"
+#define RADLDAP_MAPFILE	"modules/radldap.attrmap"
 #endif
 
 #define SEARCH_TIMEOUT	15		/* Search timeout - should be shorter
