@@ -15,6 +15,9 @@
  * 		      conf_start(). Not ideal, but still the best place IMHO.
  * 2002/03/01 - EvB - Made conf_del free interface's ACLs too
  * 2002/03/19 - EvB - Passed module base dir to subprocesses as current dir
+ * 2002/03/21 - EvB - Oops. Make that the raddb dir, otherwise legacy/users
+ * 		      etc. would have to be stored outside raddb, which would
+ * 		      not be particularly user friendly.
  */
 
 char config_id[] = "CONFIG - Copyright (C) 2001 Emile van Bergen.";
@@ -323,9 +326,9 @@ CONF *conf_new(META *m, char *basepath, char *modbasepath)
 		if (vm_getiface(vm) == ifs)
 			n = add_source(ret, m, vm->head[VM_REQ]);
 		else {
-			/* We use the module base path as prog's cwd as well */
+			/* We use raddb (basepath) as prog's cwd as well */
 			n = add_iface(ret, m, vm->head[VM_REQ], 
-				      modbasepath, modbasepath);
+				      modbasepath, basepath);
 		}
 		if (n) goto cn_cleanup;
 
