@@ -28,14 +28,12 @@ char radiusd_id[] = "RADIUSD - Copyright (C) 2001 Emile van Bergen.";
 #include <sys/select.h>
 #include <signal.h>
 #include <string.h>		/* For strerror */
-#include <malloc.h>
 #include <unistd.h>		/* For open(), dup2(), perhaps getopt() */
-#include <stdlib.h>		/* Some SysV's have getopt() here */
+#include <stdlib.h>		/* For malloc() and for some SysV's getopt() */
 #include <errno.h>
 #include <fcntl.h>		/* For O_* */
 #include <time.h>		/* For time(), time_t */
 
-
 #include <metadict.h>		/* For meta_newfromdict() */
 #include <config.h>		/* For conf_new() etc */
 #include <debug.h>		/* For msg_* */
@@ -64,9 +62,9 @@ static char *raddb = RADDB, *modbase = MODULES;
 static int nobackgr = 0;
 
 
-/* Signalling pipe */
+/* Own pid, Signalling pipe */
 
-int spipe[2] = {-1, -1};
+int pid = -1, spipe[2] = {-1, -1};
 
 
 /*
@@ -164,6 +162,9 @@ void sighandler(int sig)
 {
 	sigset_t set;
 
+	/* If we receive a signal for a forked but not execed child, exit */
+	if (getpid() != pid) _exit(125);
+
 	/* Get the current sigset from the pipe, if it's present */
 	sigemptyset(&set);
 	read(spipe[PIPE_R], &set, sizeof(set));
@@ -171,6 +172,7 @@ void sighandler(int sig)
 	/* Put the set (back) in the pipe */
 	sigaddset(&set, sig);
 	write(spipe[PIPE_W], &set, sizeof(set));
+
 	signal(sig, sighandler);
 }
 
@@ -239,6 +241,9 @@ int main(int argc, char **argv)
 		setpgid(getpid(), getpid());	/* Become pgroup leader */
 	}
 
+	/* Save pid so the sighandler can exit in not-yet-execed childs */
+	pid = getpid();
+
 	/* Bind sockets and start subprocesses */
 	time(&t); 
 	if (conf_start(c, t) == -1) { 
@@ -303,7 +308,7 @@ int main(int argc, char **argv)
 		if (sigismember(&sigs, SIGCHLD)) {
 
 			/* Handle exited childs, if any */
-			while((n = wait4(0, &status, WNOHANG, 0)) > 0) {
+			while((n = waitpid(0, &status, WNOHANG)) > 0) {
 
 				/* Find out which of our subprocesses exited */
 				for(p = c->procs; p && p->pid != n; p=p->next);
@@ -344,6 +349,8 @@ int main(int argc, char **argv)
 		/* Check if we were requested to quit */
 		if (sigismember(&sigs, SIGINT) || sigismember(&sigs, SIGTERM)) {
 
+			msg(F_PROC,L_NOTICE, "main: Got termination request\n");
+
 			/* Send all running subprocesses a SIGTERM and set
 			   reaping to the number of processes we're waiting 
 			   on to die before we exit ourselves. */
@@ -357,9 +364,8 @@ int main(int argc, char **argv)
 			/* If we don't have to wait for anybody, exit now */
 			if (reaping == 0) break;
 
-			msg(F_PROC, L_NOTICE, "main: Got termination request; "
-					      "sent SIGTERM to %d running "
-					      "childs.\n", reaping);
+			msg(F_PROC, L_NOTICE, "main: sent SIGTERM to %d "
+					      "running childs.\n", reaping);
 		}
 
 		/* Ok. Now loop through the sockets for rx readyness */
@@ -399,8 +405,14 @@ int main(int argc, char **argv)
 					chan_handle_write(p->chan, t);
 			}
 
-			/* Next, see if this proc's watchdog timed out */
-			if (p->timer && t >= p->timer) proc_handle_timeout(p,t);
+			/* Next, see if this proc's watchdog timed out. If the
+			   process has just restarted and there's work to do,
+			   kick it. */
+			if (p->timer && t >= p->timer) {
+				proc_handle_timeout(p,t);
+				if (p->state == PRS_IDLE) 
+					chan_handle_write(p->chan, t);
+			}
 		}
 
 		/* Loop through subprocs again to get new event mask. This loop
